﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Windows.Controls.Primitives;
using DlhSoft.Windows.Data;

namespace LoadChartResourceListViewDragDropSample
{
    /// <summary>
    /// Interaction logic for Window1.xaml
    /// </summary>
    public partial class Window1 : Window
    {
        public Window1()
        {
            InitializeComponent();
        }

        // The moving item thumb, source resource, and source resource item recorded when a drag operation begins.
        private Thumb itemThumb;
        private Resource sourceResource;
        private ResourceItem sourceResourceItem;

        // The initial mouse position represented by horizontal offset and vertical offset.
        private double initialHorizontalOffset, initialVerticalOffset;

        // The scroll content presenter that the drag operation may automatically scroll.
        private ScrollContentPresenter scrollContentPresenter;

        // The current vertical scroll offset updated during the drag operation.
        private double verticalScrollOffset;

        // Indicates the amount to automatically scroll when hovering near the top or bottom margin of the scroll viewport during the drag operation.
        private const double autoScrollAmount = 12;

        // Indicates the maximum distance between the mouse position and the top or bottom margin of the scroll viewport for which automatic scrolling occurs during the drag operation.
        private const double autoScrollMarginEpsilon = 8;

        // The hovering resource recorded during the drag operation.
        private Resource hoveringResource;

        // When a drag operation is started initialize the appropriate fields.
        private void ItemThumb_DragStarted(object sender, DragStartedEventArgs e)
        {
            // Record the item thumb, source resource, and source resource item.
            itemThumb = e.Source as Thumb;
            sourceResource = itemThumb != null ? itemThumb.DataContext as Resource : null;
            sourceResourceItem = sourceResource != null ? sourceResource.Item as ResourceItem : null;
            // Select the dragging resource to visually indicate the operation source.
            LoadChartResourceListView.SelectedItem = sourceResourceItem;
            // Record the initial mouse position.
            initialHorizontalOffset = e.HorizontalOffset;
            initialVerticalOffset = e.VerticalOffset;
            // Determine and record the scroll content presenter that the drag operation may automatically scroll.
            FrameworkElement sourceItemFrameworkElement = sourceResource != null ? LoadChartResourceListView.BaseEntityListView.ItemContainerGenerator.ContainerFromItem(sourceResource) as FrameworkElement : null;
            Panel parentPanel = sourceItemFrameworkElement != null ? VisualTreeHelper.GetParent(sourceItemFrameworkElement) as Panel : null;
            ItemsPresenter itemsPresenter = parentPanel != null ? VisualTreeHelper.GetParent(parentPanel) as ItemsPresenter : null;
            scrollContentPresenter = itemsPresenter != null ? VisualTreeHelper.GetParent(itemsPresenter) as ScrollContentPresenter : null;
            e.Handled = true;
        }

        // When a drag operation is in progress.
        private void ItemThumb_DragDelta(object sender, DragDeltaEventArgs e)
        {
            // Perform automatic scrolling when needed.
            ScrollViewer scrollViewer = scrollContentPresenter != null ? scrollContentPresenter.ScrollOwner : null;
            if (scrollViewer != null)
            {
                Point scrollViewerMousePosition = itemThumb.TranslatePoint(new Point(initialHorizontalOffset + e.HorizontalChange, initialVerticalOffset + e.VerticalChange), scrollViewer);
                double headerHeight = scrollViewer.ActualHeight - scrollViewer.ViewportHeight;
                double scrollOffset = 0;
                if (scrollViewerMousePosition.Y <= 0 + headerHeight + autoScrollMarginEpsilon)
                    scrollOffset = -autoScrollAmount;
                if (scrollViewerMousePosition.Y >= scrollViewer.ViewportHeight + headerHeight - autoScrollMarginEpsilon)
                    scrollOffset = +autoScrollAmount;
                double initialScrollVerticalOffset = scrollViewer.VerticalOffset;
                double updatedScrollVerticalOffset = Math.Max(Math.Min(initialScrollVerticalOffset + scrollOffset, scrollViewer.ScrollableHeight), 0);
                scrollViewer.ScrollToVerticalOffset(updatedScrollVerticalOffset);
                verticalScrollOffset += updatedScrollVerticalOffset - initialScrollVerticalOffset;
            }
            // Determine the hovering resource.
            Point mousePosition = itemThumb.TranslatePoint(new Point(initialHorizontalOffset + e.HorizontalChange, initialVerticalOffset + e.VerticalChange), scrollContentPresenter);
            HitTestResult hitTestResult = VisualTreeHelper.HitTest(scrollContentPresenter, mousePosition);
            FrameworkElement hoveringFrameworkElement = hitTestResult != null ? hitTestResult.VisualHit as FrameworkElement : null;
            hoveringResource = hoveringFrameworkElement != null ? hoveringFrameworkElement.DataContext as Resource : null;
            // Update the current vertical offset that the thumb has been dragged to in order to move or hide the horizontal line indicator in the user interface accordingly.
            FrameworkElement hoveringItemFrameworkElement = hoveringResource != null ? LoadChartResourceListView.BaseEntityListView.ItemContainerGenerator.ContainerFromItem(hoveringResource) as FrameworkElement : null;
            Point? hoveringItemFrameworkElementPosition = hoveringItemFrameworkElement != null ? (Point?)hoveringItemFrameworkElement.TranslatePoint(new Point(), itemThumb) : null;
            double? hoveringItemFrameworkElementVerticalPosition = hoveringItemFrameworkElementPosition != null ? (double?)(hoveringItemFrameworkElementPosition.Value.Y + (verticalScrollOffset + e.VerticalChange >= 0 ? hoveringItemFrameworkElement.ActualHeight : 0)) : null;
            itemThumb.Tag = hoveringItemFrameworkElementVerticalPosition != null ? (object)hoveringItemFrameworkElementVerticalPosition.Value : null;
            e.Handled = true;
        }

        // When a drag operation is completed.
        private void ItemThumb_DragCompleted(object sender, DragCompletedEventArgs e)
        {
            // Clear the current vertical offset that the thumb has been dragged to in order to hide the horizontal line indicator in the user interface.
            itemThumb.Tag = null;
            // Determine the destination resource item.
            ResourceItem destinationResourceItem = hoveringResource != null ? hoveringResource.Item as ResourceItem : null;
            if (destinationResourceItem != null && destinationResourceItem != sourceResourceItem && sourceResourceItem != null)
            {
                // Record the task assignments and start date and times to be able to reset them after moving the resource as assignments and start date and times are automatically updated when a resource is removed from the collection.
                List<KeyValuePair<Task, AssignmentItem>> taskAssignmentItems = new List<KeyValuePair<Task, AssignmentItem>>();
                Dictionary<Task, DateTime> taskStartDateTimes = new Dictionary<Task, DateTime>();
                foreach (Assignment assignment in sourceResource.Assignments)
                {
                    Task task = assignment.ParentTask;
                    taskAssignmentItems.Add(new KeyValuePair<Task, AssignmentItem>(task, assignment.Item as AssignmentItem));
                    if (taskStartDateTimes.ContainsKey(task))
                        continue;
                    taskStartDateTimes.Add(task, task.Start);
                }
                // Move the source resource item either before or after the destination resource item depending on the vertical direction of the drag operation.
                LoadChartResourceListView.Items.Remove(sourceResourceItem);
                int destinationIndex = LoadChartResourceListView.Items.IndexOf(destinationResourceItem) + (verticalScrollOffset + e.VerticalChange >= 0 ? 1 : 0);
                LoadChartResourceListView.Items.Insert(destinationIndex, sourceResourceItem);
                // Reset task assignments and start date and times.
                foreach (KeyValuePair<Task, AssignmentItem> taskAssignmentItem in taskAssignmentItems)
                {
                    Task task = taskAssignmentItem.Key;
                    AssignmentItem assignmentItem = taskAssignmentItem.Value;
                    task.Assignments.AddItem(assignmentItem);
                }
                foreach (Task task in taskStartDateTimes.Keys)
                {
                    DateTime taskStart = taskStartDateTimes[task];
                    if (taskStart != task.Start)
                        task.Start = taskStart;
                }
                LoadChartResourceListView.SelectedItem = sourceResourceItem;
            }
            // Clear fields.
            itemThumb = null;
            sourceResource = null;
            sourceResourceItem = null;
            initialHorizontalOffset = 0;
            initialVerticalOffset = 0;
            scrollContentPresenter = null;
            verticalScrollOffset = 0;
            hoveringResource = null;
            e.Handled = true;
        }
    }
}
